module Game
where

import Utility
import Maybe
import IO (Handle)
import Char (isSpace)
import Data.Set
import Data.FiniteMap
import Data.Array (Array)
import Data.Array.IArray
import Data.Array.Unboxed
import GHC.Int (Int8, Int32)

type Weight   = Int
type Currency = Int
type Coord    = Int
type Point    = (Coord, Coord)
type Metric   = Point -> Point -> Currency

memoMetric :: Game -> Metric -> Metric
memoMetric game metric =
    \pt1 pt2 -> (functions ! pt1) pt2
  where
    board     = currentBoard game
    functions = listArray (bounds board) (map metric (indices board))
		:: Array Point (Point -> Currency)

data Terrain = Plain | Water | Wall | Home deriving (Eq, Show, Read, Enum)
type PackedTerrain = Int8
packTerrain :: Terrain -> PackedTerrain
packTerrain = fromIntegral . fromEnum
unpackTerrain :: PackedTerrain -> Terrain
unpackTerrain = toEnum . fromIntegral

simplifyTerrain :: Terrain -> Terrain
simplifyTerrain Plain = Plain
simplifyTerrain Water = Wall
simplifyTerrain Wall = Wall
simplifyTerrain Home = Plain

type Board = UArray Point PackedTerrain
getSTerrain :: Board -> Point -> Terrain
getSTerrain b p = simplifyTerrain $ unpackTerrain $ b!p

newtype RobotID = RobotID Int deriving (Eq, Ord)
instance Show RobotID where
    showsPrec p (RobotID n) = showChar '#' . showsPrec p n
instance Read RobotID where
    readsPrec p str =
	[ (RobotID n, str')
	| (n, str') <- readsPrec p
	    (snd (span (\ch -> isSpace ch || ch == '#') str)) ]

data Robot = Robot
    { robotLocation :: !Point }
    deriving (Eq, Show)

newtype PackageID = PackageID Int deriving (Eq, Ord)
instance Show PackageID where
    showsPrec p (PackageID n) = showsPrec p n
instance Read PackageID where
    readsPrec p str =
	[ (PackageID n, str')
	| (n, str') <- readsPrec p str ]
instance DeepSeq PackageID where

data PackageLocation
    = Transit !RobotID			-- the robot carrying it
    | Ground !Point !(Maybe RobotID)	-- ground location and who dropped it
    deriving (Eq, Show)

data Package = Package
    { packageLocation :: !PackageLocation
	-- either which robot is carrying it, or where on the ground it is
    , packageWeight   :: !(Maybe Weight)
        -- its weight, if we know it
    , packageSource   :: !(Maybe Point)
	-- where we first noticed it being picked up or lying around, if any
    , packageTarget   :: !(Either (Set Point) Point)
	-- either the locations known to not be its target, or its known target
    } deriving (Eq, Show)

packageDelivered :: Package -> Maybe Bool
packageDelivered Package{ packageLocation = Ground loc (Just _)
                        , packageTarget   = Right tgt }
    | loc == tgt
    = Just True
packageDelivered Package{ packageLocation = Ground loc (Just _)
			, packageTarget   = Left neg }
    | not (loc `elementOf` neg)
    = Nothing
packageDelivered _
    = Just False

data Game = Game
    { serverHandle :: !Handle
    , currentBoard :: !Board
    , myRobotID    :: !RobotID
    , initCurrency :: !Currency
    , capacity     :: !Weight }
    deriving (Eq, Show)

data Status = Status
    { myCurrency :: !Currency
	-- the amount of money I have left
    , robotMap   :: FiniteMap RobotID Robot
	-- map from robot IDs to robot data
    , packageMap :: FiniteMap PackageID Package
	-- map from package IDs to package data
    , homeMap    :: FiniteMap Point Bool
	-- map from home base points to whether we have inspected each
    } deriving (Eq, Show)

myRobot :: Game -> Status -> Robot
myRobot game status = fromJust (lookupFM (robotMap status) (myRobotID game))

myRobotLocation :: Game -> Status -> Point
myRobotLocation game status = robotLocation (myRobot game status)

myRobotGain :: Game -> Status -> Weight
myRobotGain Game{ myRobotID = myRobotID }
            Status{ packageMap = packageMap }
  = sum [ wt
        | (_, Package{ packageLocation = Ground loc (Just robotID)
		     , packageWeight   = Just wt
		     , packageTarget   = Right tgt })
	    <- fmToList packageMap
	, myRobotID == robotID
	, loc == tgt ]

robotIndex :: Status -> FiniteMap Point [(RobotID, Robot)]
robotIndex Status{ robotMap = robotMap }
  = addListToFM_C prepend emptyFM
      [ (loc, [r])
      | r@(_, Robot{ robotLocation = loc })
	  <- fmToList robotMap ]

packageGroundIndex :: Status -> FiniteMap Point [(PackageID, Package)]
packageGroundIndex Status{ packageMap = packageMap }
  = addListToFM_C prepend emptyFM
      [ (loc, [p])
      | p@(_, Package{ packageLocation = Ground loc _ })
	  <- fmToList packageMap ]

packageTransitIndex :: Status -> FiniteMap RobotID [(PackageID, Package)]
packageTransitIndex Status{ packageMap = packageMap }
  = addListToFM_C prepend emptyFM
      [ (robotID, [p])
      | p@(_, Package{ packageLocation = Transit robotID })
	  <- fmToList packageMap ]

data Direction = North | South | East | West
    deriving (Eq, Show, Read, Enum)
data Command = Move !Direction | Pick [PackageID] | Drop [PackageID]
    deriving (Eq, Show, Read)
