module Geometry where

import Game
import Data.FiniteMap
import Data.Array.Unboxed
import Data.Array(Array)
import Graph
import Utility
import Maybe(maybeToList,fromJust)
import List(find,minimum)

dir :: Int -> Int -> Int
dir x y = case compare x y of
	     EQ -> 0
	     LT -> 1
	     GT -> -1

instance Num Point where
    (x1,y1) + (x2,y2) = (x1+x2,y1+y2)
    (x1,y1) - (x2,y2) = (x1-x2,y1-y2)
    negate (x,y)      = (-x,-y)
    (*) = undefined
    fromInteger = undefined
    signum = undefined
    abs = undefined

vlength :: Point -> Coord
vlength (x,y) = abs x + abs y

rotate90, rotate270, flipP :: Point -> Point
rotate90 (x,y) = (y,-x)
rotate270 (x,y) = (-y,x)
flipP (x,y) = (y,x)

-- Manhattan distance
mDist :: Metric
mDist p1 p2 = vlength (p2 - p1)

type Interval = (Coord, Coord)
inInterval :: Coord -> Interval -> Bool
inInterval x (x0, x1) = x >= x0 && x <= x1

type Chord = (Coord, Interval)
intersectChords :: Chord -> Chord -> Bool
intersectChords (y,intx) (x,inty) = inInterval y inty && inInterval x intx

data GBoard = GBoard {board :: Board,
		      gatewayLines :: GatewayLines,
		      pointGraph :: FiniteMap Point [Point]}

processBoard :: Board -> GBoard
processBoard board =
  let gls = scanBoard board
      pg = communicatingPairs board gls
  in GBoard {board = board, gatewayLines = gls, pointGraph = pg}

data ScanState = InWall
	       | AdjWall1
	       | AdjSpace1
	       | AdjWall2
	       | AdjSpace2
	       | AdjWall3
type GatewayLine1 = (Interval, [Coord])
type GatewayLine = (Coord, GatewayLine1)
type GatewayLines1 = Array Coord [GatewayLine1]
data GatewayLines = GatewayLines {horiz, vert :: GatewayLines1}
mapGLs :: (GatewayLine -> a) -> GatewayLines1 -> [a]
mapGLs f gls =
    [f (x,gl) | x <- indices gls, gl <- gls!x]

-- scanRow: Process 2 adjacent rows, looking for signature features
-- that yield gateway lines.
scanRow :: Coord -> [Terrain] -> [Terrain] -> [GatewayLine1]
scanRow x0 t1 t2 = scanRow' InWall (-1) [] x0 (zip t1 t2)

scanRow' :: ScanState -> Coord -> [Coord] -> Coord -> [(Terrain,Terrain)]
   -> [GatewayLine1]
scanRow' _      _ _ _ [] = []
scanRow' InWall _ _ x ((_,Wall):r)          = scanRow' InWall   (-1)   []   (x+1) r
scanRow' InWall _ _ x r                     = scanRow' AdjWall1  x     []     x   r
scanRow' AdjWall1  x0 _ x ((Wall,Plain):r)  = scanRow' AdjWall1  x0    []   (x+1) r
scanRow' AdjWall1  x0 _ x ((Plain,Plain):r) = scanRow' AdjSpace1 x0    []   (x+1) r
scanRow' AdjSpace1 x0 _ x ((Wall,Plain):r)  = scanRow' AdjWall2  x0  [x-1]  (x+1) r
scanRow' AdjSpace1 x0 _ x ((Plain,Plain):r) = scanRow' AdjSpace1 x0    []   (x+1) r
scanRow' AdjWall2  x0 l x ((Wall,Plain):r)  = scanRow' AdjWall2  x0    l    (x+1) r
scanRow' AdjWall2  x0 l x ((Plain,Plain):r) = scanRow' AdjSpace2 x0  (x:l)  (x+1) r
scanRow' AdjSpace2 x0 l x ((Wall, Plain):r) = scanRow' AdjWall3  x0 (x-1:l) (x+1) r
scanRow' AdjSpace2 x0 l x ((Plain, Plain):r)= scanRow' AdjSpace2 x0    l    (x+1) r
scanRow' AdjWall3 x0 l x ((Wall, Plain):r)  = scanRow' AdjWall3  x0    l    (x+1) r
scanRow' AdjWall3 x0 l x ((Plain, Plain):r) = scanRow' AdjSpace2 x0  (x:l)  (x+1) r
scanRow' AdjSpace2 x0 l x ((_, Wall):r)  =
                   ((x0,x-1),l):scanRow' InWall (-1) [] (x+1) r
scanRow' AdjWall3 x0 l x ((_, Wall):r)  =
                   ((x0,x-1),tail l):scanRow' InWall (-1) [] (x+1) r
scanRow' _ _ _ x ((_, Wall):r) = scanRow' InWall (-1) [] (x+1) r

scanBoard :: Board -> GatewayLines
scanBoard b =
  let rs = map (map (simplifyTerrain.unpackTerrain)) (rows b)
      cs = map (map (simplifyTerrain.unpackTerrain)) (cols b)
      hls1 = map2 (scanRow x0) rs
      hls2 = map2 (flip (scanRow x0)) (tail rs)
      hls = zipWith mergeLines hls1 hls2
      vls1 = map2 (scanRow y0) cs
      vls2 = map2 (flip (scanRow y0)) (tail cs)
      vls = zipWith mergeLines vls1 vls2

      mergeLines :: [GatewayLine1] -> [GatewayLine1] -> [GatewayLine1]
      mergeLines [] l2 = l2
      mergeLines l1 [] = l1
      mergeLines l1@(g1@((x1,y1),p1):l1') l2@(g2@((x2,y2),p2):l2')
	  | x1 == x2 = ((x1,y1),merge p1 p2):mergeLines l1' l2'
	  | x1 < x2 = g1:mergeLines l1 l2'
	  | x1 > x2 = g2:mergeLines l1' l2

      ((x0,y0),(x1,y1)) = bounds b
  in GatewayLines (listArray (y0+1,y1-1) hls) (listArray (x0+1,x1-1) vls)

rows, cols :: UArray (Int, Int) PackedTerrain -> [[PackedTerrain]]
rows a = [[a!(x,y) | x <- [x0..x1]] | y <- [y0..y1]]
  where ((x0,y0),(x1,y1)) = bounds a
cols a = [[a!(x,y) | y <- [y0..y1]] | x <- [x0..x1]]
  where ((x0,y0),(x1,y1)) = bounds a

ptInLines :: GatewayLines1 -> Point -> Maybe GatewayLine
ptInLines ls (x,y) = aux (ls!x)
  where aux [] = Nothing
	aux (int@((y0,y1),_):ints')
           | y < y0    = Nothing
	   | y <= y1   = Just (x,int)
	   | otherwise = aux ints'

probeLine :: Board -> Point -> Point -> [Point]
probeLine b dlt pt =
  case getSTerrain b (pt+dlt) of
   Plain -> pt:probeLine b dlt (pt+dlt)
   Wall -> let dltL = rotate90 dlt; dltR = rotate270 dlt in
     case (probeLine' b dlt dltL pt, probeLine' b dlt dltR pt) of
       ([], l) -> l
       (l, []) -> l
       _       -> undefined
probeLine' :: Board -> Point -> Point -> Point -> [Point]
probeLine' b dlt dlt' pt
   | Plain <- getSTerrain b (pt+dlt) = pt:probeLine' b dlt dlt' (pt+dlt)
   | Plain <- getSTerrain b (pt+dlt') = probeLine' b dlt dlt' (pt+dlt')
   | otherwise = []

simpleProbe :: Board -> Point -> Point -> [Point]
simpleProbe b dlt pt =
  case getSTerrain b (pt+dlt) of
    Plain -> pt:simpleProbe b dlt (pt+dlt)
    Wall -> []

sCommunicate :: Board -> Point -> Point -> Bool
sCommunicate b pt1@(x1,y1) pt2@(x2,y2)
   | x1 > x2 = sCommunicate b pt2 pt1
   | otherwise = case dropWhile (\(x,_) -> x < x2) (probeLine b pt1 (1,0)) of
		  [] -> False
		  pt3:_ -> yCommunicate b pt3 pt2

yCommunicate :: Board -> Point -> Point -> Bool
yCommunicate b pt1@(_,y1) pt2@(_,y2)
   | y1 > y2 = yCommunicate b pt2 pt1
   | otherwise = case dropWhile (\(_,y) -> y < y2) (simpleProbe b pt1 (0,1)) of
		  [] -> False
		  pt3:_ -> True

filterProbe :: GatewayLines1 -> GatewayLines1 -> [Point]
    -> ([GatewayLine], Maybe GatewayLine)
filterProbe hlines vlines pts = aux pts
  where aux [] = ([], Nothing)
	aux (p0:ps) | Just gl <- ptInLines vlines (flipP p0) = ([], Just gl)
		    | Just gl <- ptInLines hlines p0 = comb gl (aux ps)
		    | otherwise = aux ps
	comb x (xs, y) = (x:xs,y)

vhIntersect :: GatewayLine -> GatewayLine -> [(Point,Point)]
vhIntersect (x0,(_,ys)) (y0,(_,xs)) =
    [((x0,y),(x,y0)) | y <- neighbors ys y0, x <- neighbors xs x0]
vhIntersects :: Board -> GatewayLines1 -> GatewayLines1 -> GatewayLine -> [(Point,Point)]
vhIntersects board hlines vlines vl@(x0,(_,ys@(y0:_))) =
   let hls= (fst $ filterProbe hlines vlines $ probeLine board (x0,y0) (0,1))
	 ++ (fst $ filterProbe hlines vlines $ probeLine board (x0,y0) (0,-1))
   in concatMap (\hl -> vhIntersect vl hl) hls

fstPoint, lastPoint :: GatewayLine -> Coord
fstPoint  (_,(_,l)) = head l
lastPoint (_,(_,l)) = last l

hvIntersects board hlines vlines hl@(y0,(_,xs@(x0:_))) =
   let vls = (fst $ filterProbe vlines hlines $ map flipP $ probeLine board (x0,y0) (1,0))
          ++ (fst $ filterProbe vlines hlines $ map flipP $ probeLine board (x0,y0) (-1,0))
   in concatMap (\vl -> vhIntersect hl vl) vls

vvIntersects :: Board -> GatewayLines1 -> GatewayLines1 -> GatewayLine -> [(Point,Point)]
vvIntersects board hlines vlines vl0@(x0,_) =
 let proc = maybeToList . snd . filterProbe hlines vlines
     vl1 = proc $ probeLine board (x0, fstPoint vl0) (0,-1)
     vl2 = proc $ probeLine board (x0, lastPoint vl0) (0,1)
 in    map (\vl@(x,_) -> ((x0,fstPoint vl0), (x,lastPoint vl))) vl1
    ++ map (\vl@(x,_) -> ((x,fstPoint vl), (x0,lastPoint vl0))) vl2

hhIntersects :: Board -> GatewayLines1 -> GatewayLines1 -> GatewayLine -> [(Point,Point)]
hhIntersects board hlines vlines hl0@(y0,_) =
 let proc = maybeToList . snd . filterProbe vlines hlines . map flipP
     hl1 = proc $ probeLine board (fstPoint hl0, y0) (-1,0)
     hl2 = proc $ probeLine board (lastPoint hl0, y0) (1,0)
 in    map (\hl@(y,_) -> ((fstPoint hl0,y0), (lastPoint hl,y))) hl1
    ++ map (\hl@(y,_) -> ((fstPoint hl,y), (lastPoint hl0,y0))) hl2

communicatingPairs :: Board -> GatewayLines -> FiniteMap Point [Point]
communicatingPairs board (GatewayLines hlines vlines) =
    makePGraph $ concat (mapGLs (vhIntersects board hlines vlines) vlines
                      ++ mapGLs (vvIntersects board hlines vlines) vlines
                      ++ mapGLs (hhIntersects board hlines vlines) hlines)

makePGraph :: [(Point,Point)] -> FiniteMap Point [Point]
makePGraph ps = addListToFM_C (++) emptyFM
		  (map (\ (x,y) -> (x,[y])) (ps ++ map (\ (x,y)-> (y,x)) ps))

addPGraph :: FiniteMap Point [Point] -> [(Point,Point)] -> FiniteMap Point [Point]
addPGraph pg ps = addListToFM_C (++) pg
		  (map (\ (x,y) -> (x,[y])) (ps ++ map (\ (x,y)-> (y,x)) ps))

nearVertices :: GBoard -> Point -> [Point]
nearVertices (GBoard board (GatewayLines hlines vlines) pg) p0@(x0,y0) =
    let dummyHLine, dummyVLine :: GatewayLine
	dummyHLine = (y0,((x0,x0), [x0]))
	dummyVLine = (x0,((y0,y0), [y0]))
	ptPairs :: [(Point,Point)]
	ptPairs = hhIntersects board hlines vlines dummyHLine
	       ++ hvIntersects board hlines vlines dummyHLine
	       ++ vvIntersects board hlines vlines dummyVLine
	       ++ vhIntersects board hlines vlines dummyVLine
	getPt :: (Point, Point) -> Point
	getPt (p1, p2) | p0 == p1 = p2
		       | p0 == p2 = p1
    in map getPt ptPairs

boardDist :: GBoard -> Metric
boardDist gboard@(GBoard board gls pg) p0@(x0,y0) =
   let pg' = if elemFM p0 pg then pg else
          addPGraph pg [(p0,p) | p <- nearVertices gboard p0]
       bDist = fromJust . dijkstra mDist pg' p0
   in \p1 -> if elemFM p1 pg then bDist p1
	     else if sCommunicate board p0 p1 then mDist p0 p1
	     else minimum (map (\p2 -> bDist p2 + mDist p1 p2) (nearVertices gboard p1))
