{-# OPTIONS -fglasgow-exts #-}

module Knapsack
where

import Game
import Statistics
import Utility
import Maybe
import Monad
import List
import Data.FiniteMap
import Data.Array.Unboxed
import Data.Array.ST
import Data.Array.MArray
import GHC.ST
import GHC.Int (Int64)
import Control.Concurrent
import Control.Concurrent.MVar
import Control.Exception (bracket)
import CPUTime

type Epoch = Int

data Place = Place { placeBase :: !Point, placeOffset :: !Currency }
    deriving (Eq, Ord, Show)

instance DeepSeq Place where
  deepSeq (Place base offset) = deepSeq base . deepSeq offset

realPlace :: Point -> Place
realPlace pt = Place pt 0

placeMetric :: (Point -> Point -> Currency) -> (Place -> Place -> Currency)
placeMetric metric place1 place2 | place1 == place2 = 0
placeMetric metric (Place pt1 o1) (Place pt2 o2) = metric pt1 pt2 + max o1 o2

type EpochCap = (Epoch, Capability)
type PickCaps = FiniteMap Place [EpochCap]
type DropCaps = UArray Epoch Weight
type HomeCaps = FiniteMap Place [Capability]

data Capability = Capability
    { capPackageID :: !(Maybe PackageID)
    , capWeight    :: !Weight
    , capTarget    :: !Place }
    deriving (Eq, Show)

instance DeepSeq Capability where
  deepSeq (Capability p w t) = deepSeq p . deepSeq w . deepSeq t

data State = State
    { stateGain     :: !Weight
	-- my reward gained since the root of the search tree
    , stateCurrency :: !Currency
	-- my remaining amount of money
    , stateTime     :: !Int
	-- time elapsed since the root of the search tree
    , stateLocation :: !Place
	-- my current location
    , statePick     :: !PickCaps
	-- capabilities picked up, sorted by target in descending weight order
    , stateDrop     :: !DropCaps
	-- carrying commitments established by drop decisions
    , stateHomes    :: !HomeCaps }
	-- capabilities available for future pick-up
    deriving (Eq, Show)

instance DeepSeq State where
  deepSeq (State g c t l p d h)
    = deepSeq g . deepSeq c . deepSeq t
    . deepSeq l . deepSeq p . deepSeq d . deepSeq h

stateImagined :: State -> Bool
stateImagined State{ statePick = statePick }
    = any (\(_,caps) -> any (\(_,cap) -> isNothing (capPackageID cap)) caps)
          (fmToList statePick)

data Action = PickAt { actionPlace :: !Place }
	    | DropAt { actionPlace :: !Place
		     , actionDrops :: ![EpochCap] }
    deriving (Eq, Show)

instance DeepSeq Action where
    deepSeq (PickAt place    ) = deepSeq place
    deepSeq (DropAt place ecs) = deepSeq place . deepSeq ecs

data Tree = Tree !State ![(Action, Tree)]
    deriving (Eq, Show)
data Trunk = Trunk !State !Score !(Maybe (Action, Trunk, [(Action, State, Score)]))
    deriving (Eq, Show)
data Score = Score { scoreGain :: !Weight, scoreHeuristic :: !Weight }
    deriving (Eq, Show, Read)

finalState :: Trunk -> State
finalState (Trunk _ _ (Just (_, trunk, _))) = finalState trunk
finalState (Trunk state _ _) = state

trunkDrops :: Trunk -> [EpochCap] -> [EpochCap]
trunkDrops (Trunk _ _ Nothing) = id
trunkDrops (Trunk _ _ (Just (DropAt _ ecs, trunk, _))) = (ecs ++) . trunkDrops trunk
trunkDrops (Trunk _ _ (Just (_           , trunk, _))) =            trunkDrops trunk

instance DeepSeq Trunk where
  deepSeq (Trunk root score future)
    = deepSeq root . deepSeq score . deepSeq future

instance DeepSeq Score where
  deepSeq (Score g h) = deepSeq g . deepSeq h

plan :: Game -> Metric -> Status -> IO [(Action, Trunk)]
plan game metric status = do
    state0 <- randomScenario game metric status
    let onePly = expand1 game metric state0
    if null onePly then putStrLn "&&&" >> return [] else do
	let trillion = 1000000000000 :: Integer
	let psPlan0 = trillion * fromIntegral
			(minimum [ stateTime state1 | (_, state1) <- onePly ])
	branchesVar <- newMVar [ (action, trunkBy' (Tree state1 []))
			       | (action, state1) <- onePly ]
	let step depth = do
		let branches = [ (action, trunkBy' (chop depth (expand game metric state1)))
			       | (action, state1) <- onePly ]
		branches `deepSeq` swapMVar branchesVar branches
		step (depth + 1)
	startTime <- getCPUTime
	bracket
	    (forkIO (step 2))
	    killThread
	    (const
	      (let wait = do
		    now <- getCPUTime
		    branches <- readMVar branchesVar
		    let finalStates = map (finalState.snd) branches
		    let psPlan = if sizeFM (robotMap status) > 1 then psPlan0
				 else const psPlan0 $ trillion * fromIntegral
					(minimum (map stateTime finalStates))
		    let psLeft = startTime + psPlan - now
		    if psLeft > 0 && not (all stateImagined finalStates)
		       then (do threadDelay (fromIntegral (div psLeft 1000000))
				wait)
		       else (return branches)
	       in wait))
  where
    eval     = evaluate game metric status
    cmp      = compareScore 1
    trunkBy' = trunkBy eval cmp

compareScore :: Int64 -> Score -> Score -> Ordering
compareScore factor sc1 sc2
  = compare
     (fromIntegral (scoreGain sc1) + factor * fromIntegral (scoreHeuristic sc1))
     (fromIntegral (scoreGain sc2) + factor * fromIntegral (scoreHeuristic sc2))

chop :: Int -> Tree -> Tree
chop 0 (Tree root children) = Tree root []
chop n (Tree root children) = Tree root [ (action, chop (n-1) child)
					| (action, child) <- children ]

evaluate :: Game -> Metric -> Status -> State -> Score
evaluate game metric status state
    = Score (stateGain state) (heuristic game metric status state)

trunkBy :: (State -> Score) -> (Score -> Score -> Ordering) -> Tree -> Trunk
trunkBy eval cmp = trunkBy'
  where
    trunkBy' (Tree root [])
	= Trunk root (eval root) Nothing
    trunkBy' (Tree root ((carAction,carTree):childrenCdr))
	= Trunk root bestScore
	    (if stateImagined root then Nothing
	     else Just (bestAction, bestTrunk, worse))
      where
	((bestAction, bestTrunk@(Trunk _ bestScore _)), worse)
	  = f childrenCdr (carAction, trunkBy' carTree) []
	  where
	    f [] best worse = (best, worse)
	    f ((carAction, carTree) : cdr)
	      best@(bestAction, Trunk bestState bestScore _)
	      worse
	      = if cmp carScore bestScore == GT
	        then f cdr (carAction, carTrunk)
		       ((bestAction, bestState, bestScore) : worse)
		else f cdr best
		       ((carAction, carState, carScore) : worse)
	      where carTrunk@(Trunk carState carScore _) = trunkBy' carTree

heuristic :: Game -> Metric -> Status -> State -> Weight
heuristic game metric status state
  = greedy (stateCurrency state)
	   (stateDrop state)
	   (sortBy descendingPPD
	       (pickCaps ++ takeEveryNth (sizeFM (robotMap status)) homeCaps))
  where
    pickCaps
      = [ (wt, distance, Just ec)
	| (_, ecs) <- fmToList (statePick state)
	, ec@(_, Capability{ capWeight = wt, capTarget = tgt }) <- ecs
	, let distance = placeMetric metric (stateLocation state) tgt ]

    homeCaps
      = [ (wt, distance, Nothing)
	| (src, caps) <- fmToList (stateHomes state)
	, Capability{ capWeight = wt, capTarget = tgt } <- caps
	, let distance = placeMetric metric (stateLocation state) tgt ]

    takeEveryNth :: Int -> [a] -> [a]
    takeEveryNth n = f 0 where
	f _ []     = []
	f 0 (x:xs) = x:f n xs
	f n (x:xs) = f (n-1) xs

    descendingPPD :: (Weight, Currency, a) -> (Weight, Currency, a) -> Ordering
    descendingPPD (wt1, dist1, _) (wt2, dist2, _) = compare
	(fromIntegral wt2 * fromIntegral dist1 :: Int64)
	(fromIntegral wt1 * fromIntegral dist2 :: Int64)

    greedy :: Currency -> DropCaps -> [(Weight, Currency, Maybe EpochCap)] -> Weight
    greedy currency hist [] = 0
    greedy currency hist ((weight, distance, maybeEC) : rest)
      = case maybeHist' of Just hist' -> weight + greedy currency' hist' rest
			   Nothing    ->          greedy currency  hist  rest
      where currency' = currency - distance
            maybeHist' | currency' < 0     = Nothing
		       | isNothing maybeEC = Just hist
		       | otherwise         = attemptDrop game hist
						(fromJust maybeEC)

expand :: Game -> Metric -> State -> Tree
expand game metric = expand'
  where
    expand' state = Tree state [ (action, expand' state')
			       | (action, state') <- expand1' state ]
    expand1' = expand1 game metric

expand1 :: Game -> Metric -> State -> [(Action, State)]
expand1 game metric state
    = [ ( PickAt place
	, state{ stateCurrency = stateCurrency state - spending
	       , stateTime     = stateTime state + spending
	       , stateLocation = place
	       , statePick     = statePick'
	       , stateDrop     = stateDrop'
	       , stateHomes    = delFromFM (stateHomes state) place })
      | (place, caps) <- fmToList (stateHomes state)
      , place /= stateLocation state
      , let spending = placeMetric metric (stateLocation state) place
      , let newEpoch = lastEpoch + 1
      , let statePick' = addListToFM_C
			   (\ecs [ec] -> insertBy descendingWeight ec ecs)
			   (statePick state)
			   [ (capTarget cap, [(newEpoch, cap)]) | cap <- caps ]
      , let stateDrop' = listArray
			   (0, newEpoch)
			   (elems (stateDrop state) ++ [0]) ]
    ++
      [ ( DropAt place ecs
        , state{ stateGain     = stateGain state + gain
	       , stateCurrency = stateCurrency state - spending
	       , stateTime     = stateTime state + spending
	       , stateLocation = place
	       , statePick     = statePick'
	       , stateDrop     = stateDrop' } )
      | (place, ecsAvail)
	  <- fmToList (statePick state)
      , place /= stateLocation state
      , (stateDrop', ecs@(_:_), ecsAvail')
	  <- [greedy (stateDrop state) ecsAvail]
      , let spending = placeMetric metric (stateLocation state) place
      , let gain = sum [ wt | (_,Capability{capWeight=wt}) <- ecs ]
      , let statePick' = listToFM
                       $ (place, ecsAvail')
		       : [ (place', ecs'')
		         | (place', ecs') <- fmToList (statePick state)
			 , let ecs'' = filter (canDrop game stateDrop') ecs'
			 , not (null ecs'') ] ]

  where
    lastEpoch = snd (bounds (stateDrop state))

    descendingWeight (_,Capability{capWeight=wt1})
		     (_,Capability{capWeight=wt2}) = compare wt2 wt1

    greedy :: DropCaps -> [EpochCap] -> (DropCaps, [EpochCap], [EpochCap])
    greedy hist [] = (hist, [], [])
    greedy hist ecs@(ec:ecs')
      = case attemptDrop game hist ec of
	  Just hist' -> case greedy hist' ecs' of
	    (hist'', _     , []      ) -> (hist'', ecs      , []      )
	    (hist'', ecsFit, ecsUnfit) -> (hist'', ec:ecsFit, ecsUnfit)
	  Nothing -> case greedy hist ecs' of
	    (hist', []    , _       ) -> (hist', []    , ecs        )
	    (hist', ecsFit, ecsUnfit) -> (hist', ecsFit, ec:ecsUnfit)

canDrop :: Game -> DropCaps -> EpochCap -> Bool
canDrop game hist ec = isJust (attemptDrop game hist ec)

attemptDrop :: Game -> DropCaps -> EpochCap -> Maybe DropCaps
attemptDrop game hist (e0, Capability{ capWeight = capWeight })
  = runST (thaw hist >>= update e0)
  where
    lastEpoch = snd (bounds hist)

    update :: Epoch -> STUArray s Epoch Weight -> ST s (Maybe DropCaps)
    update epoch hist'
	| epoch > lastEpoch = unsafeFreeze hist' >>= return . Just
	| otherwise = do oldWt <- readArray hist' epoch
			 let newWt = oldWt + capWeight
			 if newWt > capacity game
			    then return Nothing
			    else update (epoch + 1) hist'

randomScenario :: Game -> Metric -> Status -> IO State
randomScenario game metric status = do
    homes <- homesReal emptyFM >>= homesImag
    let transitCaps = addListToFM_C prepend emptyFM
	    [ (tgtPlace, [(0, Capability (Just packageID) wt tgtPlace)])
	    | (packageID, Package{ packageLocation = Transit robotID
		                 , packageWeight   = Just wt
				 , packageTarget   = Right tgt })
		<- fmToList (packageMap status)
	    , robotID == myRobotID game
	    , let tgtPlace = realPlace tgt ]
    return State{ stateGain     = 0
		, stateCurrency = myCurrency status
		, stateTime	= 0
		, stateLocation = realPlace (myRobotLocation game status)
		, statePick	= transitCaps
		, stateDrop     = listArray (0,0) [0]
		, stateHomes    = homes }

  where
    packageGroundIndexStatus :: FiniteMap Point [(PackageID, Package)]
    packageGroundIndexStatus = packageGroundIndex status

    packageSourceCount :: FiniteMap Point Int
    packageSourceCount
      = addListToFM_C (+) emptyFM
	  [ (loc, 1)
	  | (_, Package{ packageSource = Just loc })
	      <- fmToList (packageMap status) ]

    imagineWeight :: IO Weight
    imagineWeight = imagineElt 1 x
	where x = compilePackageWeight status

    imagineTarget :: Point -> IO Place
    imagineTarget src = imagineElt 1 x >>= return . Place src
	where x = compilePackageDistance metric status

    imagineNCap :: IO Int
    imagineNCap = imagineElt 10 x
	where x = compileHomePackages game status

    packageToCap :: (PackageID, Package) -> IO Capability
    packageToCap (packageID, package) = do -- packageSource must be known
	wt  <- maybe imagineWeight return (packageWeight package)
	tgt <- case packageTarget package of
		 Left _    -> imagineTarget (fromJust (packageSource package))
		 Right tgt -> return (realPlace tgt)
	return (Capability (Just packageID) wt tgt)

    imagineCap :: Point -> IO Capability
    imagineCap src = do
	wt  <- imagineWeight
	tgt <- imagineTarget src
	return (Capability Nothing wt tgt)

    homesReal :: HomeCaps -> IO HomeCaps
    homesReal fm0 = foldFM f (return fm0) packageGroundIndexStatus
      where f loc ps mfm = do
		fm <- mfm
		let ps' = filter maybeUndelivered ps
		if null ps'
		    then return fm
		    else do caps <- mapM packageToCap ps'
			    return (addToFM_C prepend fm (realPlace loc) caps)
	    maybeUndelivered (_,pkg) = packageDelivered pkg /= Just True

    homesImag :: HomeCaps -> IO HomeCaps
    homesImag fm0 = foldFM f (return fm0) (homeMap status)
      where f loc True mfm = mfm
            f loc False mfm = do
		fm <- mfm
		ncap <- imagineNCap
		let ncap' = ncap - maybe 0 id (lookupFM packageSourceCount loc)
		if ncap' <= 0
		    then return fm
		    else do caps <- sequence (replicate ncap' (imagineCap loc))
			    return (addToFM_C prepend fm (realPlace loc) caps)

    -- TODO: maybe handle better packages with unknown delivery status
    -- TODO: metric is not really a metric (points are not zero-dist to self)
    -- TODO: package expiry

