module Main
where

import Game
import Knapsack
import Protocol
import Console
import Utility
import Data.FiniteMap
import List
import Data.Array.IArray
import Data.Array.IO
import System (getArgs)
import Char (isSpace)
import Data.Array.Unboxed
import System.IO.Unsafe (unsafePerformIO)
import Parse
import Readline
import Geometry

data Decider = Decider { run :: !(Status -> IO (Currency, Command, Decider)) }

main :: IO ()
main = do
    args <- getArgs
    let args' = if null args then ["icfp1.cse.ogi.edu", "22002"] else args
    let [serverHost, serverPort] = args'
    (game, status) <- connect serverHost (read serverPort)
    let play decider status = do
	status' <- hGetResponse game status
	printSummary game status'
	(bid, command, decider') <- run decider status'
	print (bid, command)
	hPutCommand game bid command
	play decider' status'{ myCurrency = myCurrency status - bid }
    play (decide game) status

makeMetric :: Game -> Status -> Metric
{- TODO: memoMetric -}
{-
makeMetric game status = memoMetric game m
  where
    gboard = processBoard (currentBoard game)
    m = boardDist gboard
-- makeMetric game status (x1,y1) (x2,y2) = abs(x1-x2) + abs(y1-y2)
-}
makeMetric game status = memoMetric game m
  where
    board = currentBoard game
    good pt = unpackTerrain (board!pt) `elem` [Plain, Home]
    bad = not . good

    m pt1 = unsafePerformIO $ do
	array <- newArray (bounds board) maxBound :: IO (IOUArray Point Currency)
	writeArray array pt1 0
	let process [] = return id
	    process ((pt@(x,y)):pts) = do
		dist0 <- readArray array pt
		let dist1 = dist0 + 1
		q <- mapM (\pt' ->
		    if good pt'
			then do dist <- readArray array pt'
				if dist1 < dist
				    then do writeArray array pt' dist1
					    return (pt':)
				    else return id
			else return id)
		    [ (x+1,y), (x-1,y), (x,y+1), (x,y-1) ]
		let qq = foldl1 (.) q
		qqq <- process pts
		return (qq . qqq)
	let pq [] = return ()
	    pq pts = process pts >>= \q' -> pq (q' [])
	pq [pt1]
	array' <- unsafeFreeze array :: IO (UArray Point Currency)
	return (array'!)

decide game = decider []
  where
    decider :: [(Action, Trunk)] -> Decider

    decider [] = Decider $ \status ->
      let metric = makeMetric game status in
      do alts <- plan game metric status
         if null alts
            then putStrLn "***" >> return (1, Pick [], decider [])
            else decider' metric status alts

    decider alts = Decider $ \status ->
      let metric = makeMetric game status in
      decider' metric status alts

    decider' :: Metric -> Status -> [(Action, Trunk)]
	     -> IO (Currency, Command, Decider)
    decider' metric status alts =
      let here = myRobotLocation game status in
      case find ((here ==) . placeBase . actionPlace . fst) alts of

	Just (DropAt _ ecs, trunk) ->
	  let ids = [ id
		      | (_, Capability{ capPackageID = Just id })
			  <- ecs ]
	  in if null ids then decider' metric status (trunkToAlts trunk)
		else return (1, Drop ids, decider (trunkToAlts trunk))

	Just (PickAt _, trunk) ->
	  let ids = [ id
		      | (_, Capability{ capPackageID = Just id })
			  <- trunkDrops trunk [] ]
		   ++ [ id
		      | (id, Package{ packageLocation = Ground here' _ })
			  <- fmToList (packageMap status)
		      , here == here' ]
	  in if null ids then decider' metric status (trunkToAlts trunk)
	      else return (1, Pick ids, decider (trunkToAlts trunk))

	Nothing -> do
	  direction <- move game metric status
		        [ (action, score) | (action, Trunk _ score _) <- alts ]
	  return (1, Move direction, decider alts)

move :: Game -> Metric -> Status -> [(Action, Score)] -> IO Direction
move game metric status alts
  = let bestdir = maximumBy cmp dirs
    in print dirs >> return (fst bestdir)
  where
    dirs = [ (East , evalDelta ( 1, 0))
	   , (West , evalDelta (-1, 0))
	   , (North, evalDelta ( 0, 1))
	   , (South, evalDelta ( 0,-1)) ]

    cmp (_,sc1) (_,sc2) = compare sc1 sc2

    (x0,y0) = myRobotLocation game status
    evalDelta (dx,dy) = evalPoint (x0+dx,y0+dy)
    evalPoint pt | unpackTerrain (currentBoard game ! pt) `elem` [Plain, Home]
	= maximum [ g + h + distanceToWeight
				(-metric pt (placeBase (actionPlace action)))
		  | (action, Score g h) <- alts ]
    evalPoint pt = minBound

    distanceToWeight dist
      = if den == 0 || num == 0 then dist else dist * num `div` den
      where num = myRobotGain game status
	    den = initCurrency game - myCurrency status

trunkToAlts :: Trunk -> [(Action, Trunk)]
trunkToAlts (Trunk _ _ Nothing) = []
trunkToAlts (Trunk _ _ (Just (bestAction, bestTrunk, worse))) =
    [ (bestAction, bestTrunk)
    | placeOffset (actionPlace bestAction) == 0 ]
 ++ [ (worseAction, Trunk worseState worseScore Nothing)
    | (worseAction, worseState, worseScore) <- worse
    , placeOffset (actionPlace worseAction) == 0 ]

{-
decide :: Game -> Status -> IO (Currency, Command)
decide game status = do
    printSummary game status
    let ask = do
	Just line0 <- readline " "
	addHistory line0
	let parses = [ (1  , Move dir) | (dir, line1) <- reads line0
				       , all isSpace line1 ]
		  ++ [ (1  , cmd     ) | (cmd, line1) <- reads line0
				       , all isSpace line1 ]
		  ++ [ (bid, cmd     ) | (bid, line1) <- reads line0
				       , (cmd, line2) <- reads line1
				       , all isSpace line2 ]
	case parses of [ret] -> return ret
	               []    -> putStrLn "Illegal command!" >> ask
		       _     -> putStrLn "Ambiguous command!" >> ask
    ask
-}
