module Protocol
where

import Game
import Parse
import IO
import Monad
import Maybe
import Char (isSpace)
import Network (connectTo, PortID(PortNumber))
import Data.Array.IArray
import Data.Array.IO
import Data.Set
import Data.FiniteMap

connect :: String -> Int -> IO (Game, Status)
connect serverHost serverPort = do
    handle <- connectTo serverHost (PortNumber (fromIntegral serverPort))
    hSetBuffering handle LineBuffering
    hPutStrLn handle "Player"
    hGetInit handle

hGetInit :: Handle -> IO (Game, Status)
hGetInit handle = do
    board <- hGetBoard handle
    let homeMap = listToFM [ (pt,False)
                           | (pt,x) <- assocs board
			   , unpackTerrain x == Home ]
    configLine <- hGetLine handle
    let ((myRobotID, (capacity, (myCurrency, ()))), _) = parse configLine
    return (Game handle board myRobotID myCurrency capacity,
            Status myCurrency emptyFM emptyFM homeMap)

hGetBoard :: Handle -> IO Board
hGetBoard handle = do
    sizeLine <- hGetLine handle
    let ((sizeX,(sizeY,())),_) = parse sizeLine
    mArray <- newArray ((0,0),(sizeX+1,sizeY+1)) (packTerrain Wall)
		:: IO (IOUArray Point PackedTerrain)
    mapM_ (\y -> do
	boardLine <- hGetLine handle
	zipWithM_
	    (\x ch ->
		writeArray mArray (x,y) (packTerrain (fromJust (lookup ch
		    [('.', Plain), ('~', Water), ('#', Wall), ('@', Home)]))))
	    [1..sizeX] boardLine)
	[1..sizeY]
    unsafeFreeze mArray

hGetResponse :: Game -> Status -> IO Status
hGetResponse game@Game{ serverHandle = handle, currentBoard = board } status
  = do eventsLine <- hGetLine handle
       let status' = doEvents status undefined eventsLine
       packagesLine <- hGetLine handle
       let myLocation = myRobotLocation game status'
       let status'' = doPackages myLocation status' packagesLine
       return status''

  where
    boardBounds = ((1,1),(sizeXplus1-1,sizeYplus1-1))
	    where ((0,0),(sizeXplus1,sizeYplus1)) = bounds board

    doEvents :: Status -> RobotID -> String -> Status
    doEvents status robotID str = case snd (span isSpace str) of
	""      -> status
	'#':str -> case parse str of { ((robotID',()), str) ->
	           doEvents status robotID' str }
	'E':str -> doEvents (doRelMove robotID ( 1, 0) status) robotID str
	'W':str -> doEvents (doRelMove robotID (-1, 0) status) robotID str
	'N':str -> doEvents (doRelMove robotID ( 0, 1) status) robotID str
	'S':str -> doEvents (doRelMove robotID ( 0,-1) status) robotID str
	'X':str -> case parse str of { ((x',()), str) ->
	           case snd (span isSpace str) of { 'Y':str ->
		   case parse str of { ((y',()), str) ->
		   doEvents (doAbsMove robotID (x',y') status) robotID str }}}
	'P':str -> case parse str of { ((packageID,()), str) ->
		   doEvents (doPick robotID packageID status) robotID str }
	'D':str -> case parse str of { ((packageID,()), str) ->
		   doEvents (doDrop robotID packageID status) robotID str }

    doRelMove :: RobotID -> Point -> Status -> Status
    doRelMove robotID (dx,dy) status@Status{ robotMap = robotMap } =
      case lookupFM robotMap robotID of
	Just robot@Robot{ robotLocation = (x,y) }
	  | inRange boardBounds (x+dx,y+dy)
	  -> status{ robotMap = addToFM robotMap
	      robotID robot{ robotLocation = (x+dx,y+dy) } }
	_ -> status

    doAbsMove :: RobotID -> Point -> Status -> Status
    doAbsMove robotID (x',y') status@Status{ robotMap = robotMap }
      | inRange boardBounds (x',y')
      = status{ robotMap = addToFM_C combine robotMap
	  robotID Robot{ robotLocation = (x',y') } }
      | otherwise
      = status
      where combine robot Robot{ robotLocation = loc }
                  = robot      { robotLocation = loc }

    doPick :: RobotID -> PackageID -> Status -> Status
    doPick robotID packageID
	   status@Status{ robotMap = robotMap, packageMap = packageMap }
      = status{ packageMap = addToFM_C combine packageMap packageID
	  Package{ packageLocation = Transit robotID
		 , packageWeight   = Nothing
	         , packageSource   = robotLoc
		 , packageTarget   = Left (maybe emptySet unitSet robotLoc) } }
      where
        robotLoc = lookupFM robotMap robotID >>= return . robotLocation
        combine package@Package{ packageSource   = src1
	    		       , packageTarget   = tgt1 }
                        Package{ packageLocation = loc2
	    	               , packageSource   = src2
	    	               , packageTarget   = tgt2 }
	      = package        { packageLocation = loc2
	    		       , packageSource   = src1 `mplus` src2
	    		       , packageTarget   = tgt1 `combineTgt` tgt2 }
	combineTgt tgt@(Right _) _ = tgt
	combineTgt _ tgt@(Right _) = tgt
	combineTgt (Left neg1) (Left neg2) = Left (neg1 `union` neg2)

    doDrop :: RobotID -> PackageID -> Status -> Status
    doDrop robotID packageID
	   status@Status{ robotMap = robotMap, packageMap = packageMap }
      = case lookupFM robotMap robotID of
        Just Robot{ robotLocation = robotLocation }
	  -> status{ packageMap = addToFM_C combine packageMap packageID
	       Package{ packageLocation = Ground robotLocation (Just robotID)
		      , packageWeight   = Nothing
	              , packageSource   = Nothing
		      , packageTarget   = Left emptySet } }
      where combine package Package{ packageLocation = loc }
		  = package        { packageLocation = loc }

    doPackages :: Point -> Status -> String -> Status
    doPackages here status str = case snd (span isSpace str) of
      ""  -> doPackagesDelivered here status
      str -> case parse str of { ((packageID,(tx,(ty,(wt,())))),str) ->
             doPackages here (doPackage here packageID (tx,ty) wt status) str }

    doPackage :: Point -> PackageID -> (Int,Int) -> Weight -> Status -> Status
    doPackage here packageID target wt
              status@Status{ packageMap = packageMap }
      = status{ packageMap = addToFM_C combine packageMap packageID
	  Package{ packageLocation = Ground here Nothing
	         , packageWeight   = Just wt
	         , packageSource   = Just here
	         , packageTarget   = Right target } }
      where combine package@Package{ packageLocation = loc1
                                   , packageSource   = src1 }
                            Package{ packageWeight   = wt2
				   , packageSource   = src2
				   , packageTarget   = tgt2 }
		  = package        { packageLocation = loc'
			           , packageWeight   = wt2
				   , packageSource   = src1 `mplus` src2
				   , packageTarget   = tgt2 }
		where loc' = case loc1 of Ground l d | l == here -> loc1
					  _ -> Ground here Nothing

    doPackagesDelivered :: Point -> Status -> Status
    doPackagesDelivered here status@Status{ packageMap = packageMap
					  , homeMap = homeMap }
      = status{ packageMap = mapFM f packageMap
	      , homeMap = if unpackTerrain (board!here) == Home
			  then addToFM homeMap here True else homeMap }
      where f _ package@Package{ packageLocation = Ground here' _
                               , packageTarget = Left _ }
			       | here' == here
              = package        { packageTarget = Right here }
	    f _ package
	      = package

hPutCommand :: Game -> Currency -> Command -> IO ()
hPutCommand Game{ serverHandle = handle } bid command
  = hPutStrLn handle $ show bid ++ " " ++ case command of
      Move North -> "Move N"
      Move South -> "Move S"
      Move East  -> "Move E"
      Move West  -> "Move W"
      Pick pkgs  -> "Pick" ++ concatMap ((' ':) . show) pkgs
      Drop pkgs  -> "Drop" ++ concatMap ((' ':) . show) pkgs
